/***************************************************************************
 *            process_recv.c
 *
 *  Fri Oct 24 11:28:00 2003
 *  Copyright  2003  Phil Kett
 *  pkett@genesis-midi.com
 ****************************************************************************/

#include "../include/remsync.h"
#include <sys/socket.h>
#define BUFLEN 102400
int
process_recv (char *buf, int fd)
{
	int len, pid, uid, gid;
	int inF, result, checksum, mode = 0, isdir;
	long long int bytesr, bytesw, total_read;
	off_t size;
	char command[BIG_BUF], argument[BIG_BUF], path[BIG_BUF];
	char message[BIG_BUF], amode[BIG_BUF], *buffer;
	char dummy[BIG_BUF], owner[BIG_BUF], group[BIG_BUF];
	struct stat filestat;
	bzero (message, BIG_BUF);
	bzero (command, BIG_BUF);
	bzero (argument, BIG_BUF);
	/* first of all we need to remove the trailing \r\n */
	strip_string (buf);
	len = strlen (buf);
	syslog (LOG_INFO, "Got %s\n", buf);
	/* Now we can process the string we've received */
	split_string (buf, command, argument);
	/* 
	 * we now have a command and an argument in seperate strings 
	 * now to actually make them do something!
	 * rewrite this bit to make it a switch statement
	 */
	pid = getpid ();
	if (conf.debug)
		printf ("%d: Command is '%s' '%s'\n", pid, command, argument);
	if (strcmp (command, "quit") == 0)
	{
		/*
		 * We received a quit request
		 */
		if (conf.debug)
			printf ("%d: Got Quit - returning 2\n", pid);
		return 2;
	}
	else if (strcmp (command, "csum") == 0)
	{
		int checksum = 0;
		/* checksum a file
		 * first we need to make sure that it's an ordinary file
		 */
		if (file_exists (argument) && !is_a_dir (argument))
		{
			/*
			 * We now know that the file exists and isn't a directory
			 */
			checksum = cksum (argument);
			if (conf.debug)
				printf ("%d: Checksum of %s is %d\n", pid,
					argument, checksum);
			/* Construct reply and send over network */
			sprintf (message, "CSUM:%s:%d\r\n", argument,
				 checksum);
			send (fd, message, strlen (message), 0);
		}
		else
		{
			sprintf (message, "Error: Cannot checksum %s\r\n",
				 argument);
			send (fd, message, strlen (message), 0);
		}
	}
	else if (strcmp (command, "scan") == 0)
	{
		/*
		 * Scan a directory structure and return the checksums
		 */
		if (is_a_dir (argument))
		{
			syslog (LOG_INFO, "Initiating scan of %s\n",
				argument);
			net_dir_scan (fd, argument, 0);
			sprintf (message, "DONE\r\n");
			send (fd, message, strlen (message), 0);
		}
		else
		{
			sprintf (message,
				 "Error: Cannot scan %s: Not a directory\r\n",
				 argument);
			send (fd, message, strlen (message), 0);
		}
	}
	else if (strcmp (command, "send") == 0)
	{
		/*
		 * Send file over the network
		 */

		/*
		 * first check the argument we've been given is valid
		 */
		if (is_a_dir (argument))
		{
			sprintf (message,
				 "Error: Cannot send %s: Not an ordinary file\r\n",
				 argument);
			send (fd, message, strlen (message), 0);
		}
		else
		{
			/* 
			 * Now we send the file over the network
			 */
			/* 
			 * do a stat to get file sizes and a checksum we can send
			 */
			result = stat (argument, &filestat);
			if (result != 0)
			{
				sprintf (message,
					 "Error: Cannot stat file %s\r\n",
					 argument);
				send (fd, message, strlen (message), 0);
				if (conf.debug) printf("%s", message);
				return 2;
			}
			size = filestat.st_size;
			checksum = cksum (argument);
			bzero (message, BIG_BUF);
			if (conf.debug) syslog (LOG_INFO, "%d: Sending file %s\n", pid,
				argument);
			sprintf (message, "Sending:%s:%Ld:%d\r\n", argument,
				 size, checksum);
			send (fd, message, strlen (message), 0);

			do
			{
				bzero (message, BIG_BUF);
				read (fd, message, BIG_BUF);
				strip_string (message);
			}
			while (strcmp (message, "OK") != 0);
			if (conf.debug) printf("Reading %Ld bytes\n", size);
			
			inF = open (argument, O_RDONLY | O_BINARY);
			total_read = 0;
			
			buffer = malloc (BUFLEN);
			while (total_read < size) {
			bzero (buffer, BUFLEN);
			bytesr = read (inF, buffer, BUFLEN);
			
			bytesw = write (fd, buffer, bytesr);
			total_read += bytesr;
			}
			free (buffer);

			close (inF);
		}
	}
	else if (strcmp (command, "recv") == 0)
	{
		if (conf.debug) syslog (LOG_INFO, "%d: Got recv command %s\n", pid, argument);
		/* Receive file from client */
		result = server_recv_file (argument, fd);
	}
	else if (strcmp (command, "mdir") == 0)
	{
		/* Make a directory */
		/* first of all split out the mode and the path */
		if (conf.debug) logmsg ("Got mdir command");
		split_string (argument, path, amode);
		mode = atoi (amode);
		if (conf.debug) syslog (LOG_INFO, "About to run mkdir(%s, %d)\n", path, mode);
		result = mkdir (path, mode);
		if (result != 0)
		{
			logmsg ("Mkdir failed\n");
		}
		bzero (message, BIG_BUF);
		sprintf (message, "OK\r\n");
		send (fd, message, strlen (message), 0);
	}
	else if (strcmp (command, "dele") == 0)
	{
		/* Delete a file */
		/* first we need to check if this is a directory */
		isdir = is_a_dir (argument);
		if (isdir == 1)
		{
			/* We have a directory to deal with */
			checkslash (argument);
			dirdelete (argument);
			bzero (message, BIG_BUF);
			sprintf (message, "OK\r\n");
			send (fd, message, strlen (message), 0);
		}
		else
		{
			/* It's a file */
			result = unlink (argument);
			if (result != 0)
			{
				logmsg ("Unlink failed");
			}
			bzero (message, BIG_BUF);
			sprintf (message, "OK\r\n");
			send (fd, message, strlen (message), 0);
		}
	}
	else if (strcmp (command, "stat") == 0)
	{

	}
	else if (strcmp (command, "exst") == 0)
	{
		if (conf.debug) printf("exst command: Checking for %s\n", argument);
		/* checking if file exists */
		if (file_exists (argument))
		{
			if (conf.debug) printf("YES\n");
			bzero (message, BIG_BUF);
			sprintf (message, "YES\r\n");
			send (fd, message, strlen (message), 0);
		}
		else
		{
			if (conf.debug) printf("NO\n");
			bzero (message, BIG_BUF);
			sprintf (message, "NO\r\n");
			send (fd, message, strlen (message), 0);
		}
	}

	else if (strcmp (command, "cmod") == 0)
	{
		/* Chmod command received */
		if (conf.debug) logmsg ("Got chmod command\n");
		split_string (argument, path, amode);
		mode = atoi (mode);
		result = chmod (path, mode);
		if (result != 0)
		{
			logmsg ("chmod failed\n");
		}
		bzero (message, BIG_BUF);
		sprintf (message, "OK\r\n");
		send (fd, message, strlen (message), 0);
	}
	else if (strcmp (command, "cown") == 0)
	{
		/* Chown command received */
		if (conf.debug) logmsg ("Got Chown command\n");
		split_string (argument, path, dummy);
		split_string (dummy, owner, group);
		uid = atoi (owner);
		gid = atoi (group);
		result = chown (path, owner, group);
		if (result != 0)
		{
			logmsg ("Chown Failed\n");
		}
		bzero (message, BIG_BUF);
		sprintf (message, "OK\r\n");
		send (fd, message, strlen (message), 0);
	}
	if (conf.debug)
		printf ("Command '%s'\n", command);
	if (conf.debug)
		printf ("Argument '%s'\n", argument);

	return 1;
}
