/***************************************************************************
 *            set_monitor.c
 *
 *  Wed Jan  7 15:40:44 2004
 *  Copyright  2004  Phil Kett
 *  pkett@genesis-midi.com
 ****************************************************************************/
#include "../include/remsync.h"

#ifdef HAVE_FNOTIFY

static volatile int event_fd1;
static volatile int event_fd2;
static volatile int event_fd3;
static volatile int got_kill;

void sigkill_handler (int s) 
{
	got_kill = 1;
}

void handler1(int sig, siginfo_t *si, void *data)
{
    /* Handler for Modify events */
    event_fd1 = si->si_fd;
}

void handler2(int sig, siginfo_t *si, void *data)
{
    /* Handler for create events */
    event_fd2 = si->si_fd;
}

void handler3(int sig, siginfo_t *si, void *data)
{
    /* Handler for Delete Events */
    event_fd3 = si->si_fd;
}



int set_monitor (char *directory) 
{
    /*
     * Set a filesystem monitor on a directory tree
     * spawning a new process for each subdir
     *
     * This also stores a list of each child process so that
     * it can kill processes whose directories have been deleted
     * (linked list hell....)
     */

    /* Structures for lists */
    struct dir_list 
    {
        struct dir_list *next_dir_ptr;
        int isdir;
        time_t mod_time;
        char filename[BIG_BUF];
        pid_t proc_id;
    };

    struct dir_list *first_dir_ptr = NULL;
    struct dir_list *new_dir_entry;
    struct dir_list *current_ptr;
    struct dir_list *prev_ptr;
    struct dir_list *lose_ptr;

    /* Structures for Signal Handlers */
    struct sigaction act1;
    struct sigaction act2;
    struct sigaction act3;
	struct sigaction sa_chld;
	struct sigaction sa_kill;
	

    /* File descriptors for directory monitor */
    int fd1, fd2, fd3;

    /* Structure for File stats */
    struct stat filestat;

    
    struct dirent *d;
    DIR *dir;
    int isdir, tmp, notify = 0, len, count;
    pid_t pid;
    int found, result, loop;
    char fulldir[BIG_BUF], fulldest[BIG_BUF], notifyname[BIG_BUF], fullnotifyname[BIG_BUF];

	/*sigkill handler (Cleans up child processes) */
	sa_kill.sa_handler = sigkill_handler;
	sigemptyset (&sa_kill.sa_mask);
	sa_kill.sa_flags = SA_RESTART;
	if (sigaction (SIGTERM, &sa_kill, NULL) == -1)
	{
		logmsg("sigaction term");
		exit (1);
	}

	
	/* Sigchld handler (stops zombie processes) */
	sa_chld.sa_handler = sigchld_handler;
	sigemptyset (&sa_chld.sa_mask);
	sa_chld.sa_flags = SA_RESTART;
	if (sigaction (SIGCHLD, &sa_chld, NULL) == -1)
	{
		logmsg("sigaction chld");
		exit (1);
	}
    if (conf.debug) printf("Setting Monitor on %s\n", directory);
    
	syslog(LOG_INFO, "Setting Monitor on %s\n", directory);
    /* set up Sig handler for create events */
    act2.sa_sigaction = handler2;
    sigemptyset(&act2.sa_mask);
    act2.sa_flags = SA_SIGINFO;
    sigaction(SIGRTMIN+3, &act2, NULL);

    /* set up Sig handler for Modify events */
    act1.sa_sigaction = handler1;
    sigemptyset(&act1.sa_mask);
    act1.sa_flags = SA_SIGINFO;
    sigaction(SIGRTMIN+5, &act1, NULL);

    /* Set up Sig handler for Delete events */
    act3.sa_sigaction = handler3;
    sigemptyset(&act3.sa_mask);
    act3.sa_flags = SA_SIGINFO;
    sigaction(SIGRTMIN+4, &act3, NULL);


    /* open directories and set up monitors */
    fd1 = open(directory, O_RDONLY);
    fcntl(fd1, F_SETSIG, SIGRTMIN+3);
    fcntl(fd1, F_NOTIFY, DN_CREATE|DN_MULTISHOT);
    fd2 = open(directory, O_RDONLY);
    fcntl(fd2, F_SETSIG, SIGRTMIN+5);
    fcntl(fd2, F_NOTIFY, DN_RENAME|DN_ATTRIB|DN_MODIFY|DN_MULTISHOT);
    fd3 = open(directory, O_RDONLY);
    fcntl(fd3, F_SETSIG, SIGRTMIN+4);
    fcntl(fd3, F_NOTIFY, DN_DELETE|DN_MULTISHOT);

    /*
     * Do directory scan and store in linked list
     * this is so that later we can try and identify which file has changed
     */
    dir = opendir(directory);
    //printf("Directory open\n");
    
    if (dir == NULL) {
        perror("Failed");
        printf("Failed\n");
        
        exit (-1);
    }
    //printf("Entering Loop\n");
    while ((d=readdir(dir))) {
        if (conf.debug) printf("Inside Loop\n");
        if (conf.debug) printf("Processing file %s\n", d->d_name);
        if (strcmp(d->d_name, ".") != 0 && strcmp(d->d_name, "..") !=0) {
            if (conf.debug) printf("Assigning memory\n");
            new_dir_entry = malloc(sizeof(struct dir_list));
            //printf("Strcpy\n");
            snprintf(fulldir, BIG_BUF, "%s%s", directory, d->d_name);
            //printf("%s\n", fulldir);
            tmp = stat(fulldir, &filestat);
            if (tmp != 0) {
                printf("Stat of %s Failed\n", d->d_name);
            }
            new_dir_entry->mod_time = filestat.st_ctime;
            strncpy(new_dir_entry->filename, d->d_name, BIG_BUF);
            /* Check if this particular entry is a directory
             * If it is we spawn a new scan for this entry
             * and add the relevant info to the structure
             * ->isdir=1 and ->proc_id=pid
             */
            
            isdir = is_a_dir(fulldir);
            if (isdir == 1) {
                checkslash(fulldir);
                new_dir_entry->isdir = isdir;
                if ((pid=fork()) == 0) {
                    set_monitor(fulldir);
                }
                new_dir_entry->proc_id=pid;
            } else {
                new_dir_entry->isdir = 0;
                new_dir_entry->proc_id = 0;
            }
            //printf("Assigning pointers\n");
            new_dir_entry->next_dir_ptr = first_dir_ptr;
            //printf("Setting up next in chain\n");
            first_dir_ptr = new_dir_entry;
        }
    }
    closedir(dir);
    //printf("Count: %d\n", count);

	/* Check to see if we're using the notify system */
	if (strlen(conf.notify) != 0) {
		notify = 1;
	}
	
    /* Now for the real work */
    while (1) {
        //printf("Waiting for signals\n");
        
        event_fd1 = (int)NULL;
        event_fd2 = (int)NULL;
        event_fd3 = (int)NULL;
		got_kill = (int)NULL;
        pause();
		if (got_kill) {
			/* Process has received a sigkill
			* clean up all children and exit
			*/
			syslog(LOG_INFO, "Got Kill signal... Exiting\n");
			current_ptr = first_dir_ptr;
            prev_ptr = NULL;
            while (current_ptr != NULL) {
				if (current_ptr->isdir == 1) {
                    /* it was a directory so we need to kill the monitor process and
                    * remove it from the list of monitored directories
                    */
                    //printf("Killing Process %d\n", current_ptr->proc_id);
                    kill(current_ptr->proc_id, SIGTERM);
                    /* Process killed */
                
					/* Check to see if this is the first entry in the list */
					if (prev_ptr != NULL) {
						prev_ptr->next_dir_ptr = current_ptr->next_dir_ptr;
					} else {
						first_dir_ptr = current_ptr->next_dir_ptr;
					}
                    
					lose_ptr = current_ptr;
					current_ptr = current_ptr->next_dir_ptr;
					//printf("About to lose %s\n", lose_ptr->filename);
					/* Free the memory */
					free(lose_ptr);
				} else {
						prev_ptr = current_ptr;
						current_ptr = current_ptr->next_dir_ptr;
				}
            }
			syslog(LOG_INFO,"Cleanup complete - exiting");
			exit(0);
		}
        if (event_fd2) {
			if (conf.debug) syslog(LOG_INFO, "Got Create event on fd %d directory %s\n", event_fd2, directory);
            //printf("Got Create event on fd %d directory %s\n", event_fd2, directory);
            /*
             * find out what's been created!
             */
            dir = opendir(directory);
            while ((d=readdir(dir))) {
                if (strcmp(d->d_name, ".") != 0 && strcmp(d->d_name, "..") !=0 && strstr(d->d_name, "remsync") == 0) {
                    //printf("Running dirfind\n");

                    //printf("Setting current_ptr\n");
                
                    current_ptr = first_dir_ptr;
                
                    while ((current_ptr != NULL)) {
                        if (strncmp(current_ptr->filename, d->d_name, BIG_BUF) == 0) {
                            break;
                        }
                        
                        //printf("Inside loop %s %s\n", current_ptr->filename, d->d_name);
                        
                        current_ptr = current_ptr->next_dir_ptr;
                    }
                    if (current_ptr == NULL) {
                        //printf("%s was created\n", d->d_name);
                        /* Update the linked list to add the new entry */
                        new_dir_entry = malloc(sizeof(struct dir_list));
                        snprintf(fulldir, BIG_BUF, "%s%s", directory, d->d_name);
                        tmp = stat(fulldir, &filestat);
                        new_dir_entry->mod_time = filestat.st_ctime;
                        strncpy(new_dir_entry->filename, d->d_name, BIG_BUF);
                        /* Do we have a new directory? */
						if ((notify == 1)) {
							/* we're using the notify system 
							* Work out what notifyname is
							*/
							len = strlen (fulldir);
							for (count = 0; count <= len; count++)
							{
								if (fulldir[count] == '/')
								{
									notifyname[count] = '~';
								}
								else
								{
									notifyname[count] = fulldir[count];
								}
							}
							notifyname[len+1] = '\0';
							sprintf(fullnotifyname, "%s%s", conf.notify, notifyname);
						}
                        isdir = is_a_dir(fulldir);
                        if (isdir == 1) {
                            /* Yes we do! */
                            new_dir_entry->isdir = isdir;
                            checkslash(fulldir);
                            if ((pid = fork()) == 0) {
                                set_monitor(fulldir);
                            }
                            new_dir_entry->proc_id = pid;
							/* Send command to server to create new directory */
							if ((notify = 1) && (file_exists(fullnotifyname))) {
								/* Skip this file - we've been notified */
								unlink(fullnotifyname);
							} else {
							server_mdir(fulldir);
							}
                        } else {
                            new_dir_entry->isdir = 0;
                            new_dir_entry->proc_id = 0;
							/* Send new file over */
							if ((notify = 1) && (file_exists(fullnotifyname))) {
								/* Skip this file - we've been notified */
								unlink(fullnotifyname);
							} else {
								for (loop=0; loop<5; loop++) {
									if (conf.debug) syslog(LOG_INFO, "Sending over file %s\n", fulldir);
									/* Try to copy over the file a maximum of 5 times */
									result = client_send_file(fulldir);
									if (result == 0) {
										break;
									}
								}
							}
                        }
                        //printf("Directory? %d\n", new_dir_entry->isdir);
                        new_dir_entry->next_dir_ptr = first_dir_ptr;
                        first_dir_ptr = new_dir_entry;
                        break;
                    }
                }
            }
            closedir(dir);
        }
    
        if (event_fd1) {
            if (conf.debug) printf("Got Modify event on fd %d directory %s\n", event_fd1, directory);
            /* Find out what's changed */
            dir = opendir(directory);
            while ((d=readdir(dir))) {
				if (strcmp(d->d_name, ".") != 0 && strcmp(d->d_name, "..") !=0 && strstr(d->d_name, "remsync") == 0) {
                    current_ptr = first_dir_ptr;
                    while ((current_ptr != NULL)) {
                        /* Scan through list until we find the current filename */
                        if (strcmp(current_ptr->filename, d->d_name) == 0) {
                            snprintf(fulldir, BIG_BUF, "%s%s", directory, d->d_name);
                            tmp = lstat(fulldir, &filestat);
                            if (filestat.st_ctime != current_ptr->mod_time) {
                                /* We've found a file that's been modified - update the list */
                                current_ptr->mod_time = filestat.st_ctime;
								if ((notify == 1)) {
									/* we're using the notify system 
									* Work out what notifyname is
									*/
									len = strlen (fulldir);
									for (count = 0; count <= len; count++)
									{
										if (fulldir[count] == '/')
										{
											notifyname[count] = '~';
										}
										else
										{
											notifyname[count] = fulldir[count];
										}
									}
									notifyname[len+1] = '\0';
									sprintf(fullnotifyname, "%s%s", conf.notify, notifyname);
								}
								if (is_a_dir(fulldir)) {
									/* This is a directory - we only support chmod changes on directories */
									server_cmod(fulldir);
								} else {
									if (conf.debug) syslog(LOG_INFO, "Checking for %s\n", fullnotifyname);
									if ((notify = 1) && (file_exists(fullnotifyname))) {
									/* Skip this file - we've been notified */
									unlink(fullnotifyname);
									} else {
										/* It's an ordinary file - can be copied over */
										/* Now send the modified file over */
										for (loop=0; loop<5; loop++) {
											/* Try to copy over the file a maximum of 5 times */
											result = client_send_file(fulldir);
											if (result == 0) {
												break;
											}
										}
									}
								}
                                break;
                            }
                        }
                        current_ptr = current_ptr->next_dir_ptr;
                    }
                }
            }
        }
        if (event_fd3) {
            if (conf.debug) printf("Got Delete event on fd %d directory %s\n", event_fd1, directory);
            current_ptr = first_dir_ptr;
            prev_ptr = NULL;
            while (current_ptr != NULL) {
                /* Scan through the linked list to try and find a file that no longer exists */
                dir = opendir(directory);
                found = 0;
                while ((d = readdir(dir))) {
                    if (strcmp(current_ptr->filename, d->d_name) == 0) {
                        /* This file still exists  so set a flag */
                        found = 1;
                        break;
                        
                    }
                }
                closedir(dir);
                 if (found != 1) {
                    /* This file doesn't exist - it must have been deleted */
                    //printf("%s was deleted\n", current_ptr->filename);
					 sprintf(fulldir, "%s%s", directory, current_ptr->filename);
					 conv_src_to_dest(conf.source, conf.dest, fulldir, fulldest);
					 server_dele(fulldest);
                    /* remove entry from list */
                    /* Before we remove the entry from the list we need to check to see if it was a directory */
                    if (current_ptr->isdir == 1) {
                        /* it was a directory so we need to kill the monitor process and
                         * remove it from the list of monitored directories
                         */
                        //printf("Killing Process %d\n", current_ptr->proc_id);
                        kill(current_ptr->proc_id, SIGKILL);
                        /* Process killed */
                    }
                    /* Check to see if this is the first entry in the list */
                    if (prev_ptr != NULL) {
                        prev_ptr->next_dir_ptr = current_ptr->next_dir_ptr;
                    } else {
                        first_dir_ptr = current_ptr->next_dir_ptr;
                    }
                    
                    lose_ptr = current_ptr;
                    current_ptr = current_ptr->next_dir_ptr;
                    //printf("About to lose %s\n", lose_ptr->filename);
                    /* Free the memory */
                    free(lose_ptr);
                } else {
                    prev_ptr = current_ptr;
                    current_ptr = current_ptr->next_dir_ptr;
                }
            }
            
        }
    }
    
}
#endif
